@@ -1,4 +1,5 @@
 from abc import ABC, abstractmethod, abstractproperty, abstractstaticmethod
+from inspect import isfunction
 from io import BytesIO as _BytesIO
 import re as _re
 from datetime import datetime, timedelta
@@ -8,7 +9,9 @@ from pathlib import Path
 from typing import (
     TYPE_CHECKING,
     BinaryIO,
+    Callable,
     Iterable,
+    Iterator,
     Literal,
     NamedTuple,
     Tuple,
@@ -45,14 +48,27 @@ from cryptography.x509 import Certificate, CertificateBuilder
 from cryptography.x509.oid import ExtendedKeyUsageOID
 from typing_extensions import TypeGuard as _TypeGuard
 from os import PathLike
+import tarfile as _tar
 
 if TYPE_CHECKING:
     from OpenSSL import crypto as _crypto
+    from sslcontext import SSLContext, is_pyopenssl
 else:
     try:
-        from OpenSSL import crypto as _crypto
+        from OpenSSL import crypto as _crypto, SSL as _SSL
+        try:
+            from sslcontext import is_pyopenssl
+        except:
+            def is_pyopenssl(ctx):
+                if hasattr(ctx, "_ctx") and isinstance(ctx._ctx, _SSL.Context):
+                    return True
+                else:
+                    return False
+
     except ImportError:
         _crypto = None
+        def is_pyopenssl(ctx):
+            return False
 
 ExtensionLike = Union[
     x509.Extension,
@@ -105,7 +121,7 @@ class Encoding(_IntEnum):
 
 
 ValidPath = Union[str, PathLike]
-PasswordLike = Union[bytes, str, None]
+PasswordLike = Union[bytes, str, None, Callable[[], Union[str, bytes]]]
 
 ProtectedByteEncoded = Tuple[bytes, Encoding, PasswordLike]
 EncodedBytes = Union[Tuple[bytes, Encoding], ProtectedByteEncoded]
@@ -121,16 +137,21 @@ def _is_encoded_file(encoded: Encoded) -> _TypeGuard[EncodedFile]:
 
 
 def _getEncryption(password: PasswordLike) -> _Encryption:
-    if isinstance(password, _Encryption):
-        return password
-    elif isinstance(password, str):
-        return _TextEncryption(password.encode())
-    elif password is None:
+    password = _getPassword(password)
+    if password is None:
         return _NoEncryption()
     else:
         return _TextEncryption(password)
 
 
+def _getPassword(password: PasswordLike) -> "bytes|None":
+    if password is None:
+        return None
+    elif isfunction(password):
+        password = password()
+    return _as_bytes(password) if password is not None else None
+
+
 def _as_bytes(val, encoding: str = "utf-8") -> bytes:
     return val.encode(encoding) if isinstance(val, str) else val
 
@@ -172,8 +193,13 @@ class X509EncodedStore(ABC):
     def __exit__(self, *args):
         self.close()
 
-    def decode(self):
-        return load_store(*self.raw_dump())
+    def iter_decode(self):
+        with self as (io, encoding, password):
+            return load_all(io, encoding, password)
+
+    def load_key_and_certificates(self):
+        with self as (io, encoding, password):
+            return load_key_and_certificates(io, encoding, password)
 
 
 class __ByteDecoder(NamedTuple):
@@ -231,13 +257,13 @@ class __FileDecoder(NamedTuple):
             encoding = Encoding.from_suffix(path.suffix)
         elif isinstance(encoding, str):
             encoding = Encoding[encoding]
-    
+
         return _FileDecoder(path, encoding, encryption)
 
     def open(self, mode: str = "rb"):
         self._io = self.path.open(mode)
         return (self._io, self.encoding, self.encryption)
-    
+
     def close(self):
         return self._io.close()
 
@@ -263,8 +289,44 @@ class _X509PubCreds(NamedTuple):
     chain: "list[Certificate]"
 
 
+def load_der(data: bytes, password: bytes = None):
+    try:
+        return x509.load_der_x509_certificate(data)
+    except ValueError:
+        return _load_der_private_key(data, password)
+
+
+def dump_der_archive(file: BinaryIO, store: "Iterable[Certificate|PrivateKey]"):
+    with _tar.open(fileobj=file, mode="w") as bundle:
+        for i, cert_or_key in enumerate(store):
+            if isinstance(cert_or_key, Certificate):
+                data = dump_cert(cert_or_key, Encoding.DER)
+                suffix = "crt"
+            else:
+                data = dump_key(cert_or_key, Encoding.DER)
+                suffix = "key"
+            info = _tar.TarInfo(f"{i}.{suffix}.der")
+            info.size = len(data)
+            bundle.addfile(info, _BytesIO(data))
+
+
+def load_der_archive(file: "BinaryIO|_tar.TarFile|bytes", password: bytes = None):
+    def _load(file: _tar.TarFile):
+        for member in file.getmembers():
+            if member.name.endswith(".der"):
+                yield load_der(file.extractfile(member).read(), password)
+
+    if isinstance(file, _tar.TarFile):
+        yield from _load(file)
+    else:
+        with _tar.open(
+            fileobj=file if isinstance(file, bytes) else file, mode="r"
+        ) as file:
+            yield from _load(file)
+
+
 def load_cert(data: bytes, encoding: Encoding, password: PasswordLike = None):
-    password = _as_bytes(password)
+    password = _getPassword(password)
     if encoding is Encoding.PKCS12:
         store = _pkcs12.load_pkcs12(data, password)
         return store.cert.certificate
@@ -277,7 +339,7 @@ def load_cert(data: bytes, encoding: Encoding, password: PasswordLike = None):
 
 
 def load_key(data: bytes, encoding: Encoding, password: PasswordLike = None):
-    password = _as_bytes(password)
+    password = _getPassword(password)
     if encoding is Encoding.PKCS12:
         store = _pkcs12.load_pkcs12(data, password)
         if store.key is None:
@@ -329,71 +391,102 @@ def dump_key(key: PrivateKey, encoding: Encoding, password: PasswordLike = None)
 _PEM_SECTION = _re.compile(rb"-----([^- ]+) ([^-]+)-----")
 
 
-def parse_pem(pem: bytes):
-    result: "list[tuple[str, bytes]]" = []
+def iter_pem(pem: BinaryIO) -> Iterator[Tuple[str, bytes]]:
     sectionName = None
     sectionData = bytes()
-    for line in pem.splitlines():
+    while True:
+        line = pem.readline()
+        if not line:
+            break
         line = line.strip()
         match = _PEM_SECTION.match(line)
+        sectionData += b"\n" + line
         if match:
-            sectionData += b"\n" + line
             if sectionName is not None:
                 expecting = b"-----END " + sectionName + b"-----"
                 if expecting != match[0]:
                     raise ValueError(f"Expecting: {expecting} Found:{match[0]}")
-                result.append((sectionName.decode(), sectionData))
+                yield sectionName.decode(), sectionData
                 sectionName = None
                 sectionData = bytes()
             else:
                 if match[1] != b"BEGIN":
                     raise ValueError(f"Expecting: BEGIN Found:{match[0]}")
                 sectionName = match[2]
-        elif sectionName:
-            sectionData += b"\n" + line
-    return result
+
+
+def decode_pem(pem: "BinaryIO|bytes", password: PasswordLike = None):
+    for section, data in iter_pem(_BytesIO(pem) if isinstance(pem, bytes) else pem):
+        if section == "CERTIFICATE":
+            yield x509.load_pem_x509_certificate(data)
+        elif "PRIVATE KEY" in section:
+            yield _load_pem_private_key(data, password)
 
 
 def load_certs(
-    data: bytes, encoding: Encoding, password: PasswordLike = None
+    io: "bytes|BinaryIO", encoding: Encoding, password: PasswordLike = None
 ) -> "list[Certificate]":
-    password = _as_bytes(password)
+    password = _getPassword(password)
     if encoding is Encoding.PKCS12:
-        store = _pkcs12.load_pkcs12(data, password)
+        store = _pkcs12.load_pkcs12(
+            io.read() if not isinstance(io, bytes) else io, password
+        )
         return [store.cert.certificate] + store.additional_certs
     elif encoding is Encoding.PEM:
         certs = []
-        for section, data in parse_pem(data):
+        for section, data in iter_pem(io):
             if section == "CERTIFICATE":
                 certs.append(x509.load_pem_x509_certificate(data))
         return certs
     elif encoding is Encoding.DER:
-        return [x509.load_der_x509_certificate(data)]
+        return [
+            x509.load_der_x509_certificate(
+                io.read() if not isinstance(io, bytes) else io
+            )
+        ]
     else:
         raise ValueError(f"Invalid encoding {encoding}")
 
 
-def load_store(
-    data: bytes, encoding: Encoding, password: PasswordLike = None
+def load_key_and_certificates(
+    io: "bytes|BinaryIO", encoding: Encoding, password: PasswordLike = None
 ) -> "tuple[Certificate|None, PrivateKey|None, list[Certificate]]":
-    password = _as_bytes(password)
+    password = _getPassword(password)
     if encoding is Encoding.PKCS12:
-        store = _pkcs12.load_pkcs12(data, password)
-        return store.cert.certificate, store.key, store.additional_certs
+        key, cert, chain = _pkcs12.load_key_and_certificates(
+            io.read() if not isinstance(io, bytes) else io, password
+        )
+        return cert, key, chain
     elif encoding is Encoding.PEM:
         key = None
         certs = []
-        for section, data in parse_pem(data):
+        for section, data in iter_pem(io):
             if section == "CERTIFICATE":
                 certs.append(x509.load_pem_x509_certificate(data))
             elif "PRIVATE KEY" in section and key is None:
                 key = _load_pem_private_key(data, password)
         return certs[0], key, certs[1:]
     elif encoding is Encoding.DER:
-        try:
-            return x509.load_der_x509_certificate(data), None, []
-        except ValueError:
-            return None, _load_der_private_key(data, password), []
+        return load_der(io.read() if not isinstance(io, bytes) else io, password)
+    else:
+        raise ValueError(f"Invalid encoding {encoding}")
+
+
+def load_all(
+    io: "BinaryIO|bytes", encoding: Encoding, password: PasswordLike = None
+) -> "Iterator[PrivateKey|Certificate]":
+    password = _getPassword(password)
+    if encoding is Encoding.PKCS12:
+        key, cert, chain = _pkcs12.load_key_and_certificates(
+            io.read() if not isinstance(io, bytes) else io, password
+        )
+        yield cert
+        yield key
+        yield from chain
+    elif encoding is Encoding.PEM:
+        yield from decode_pem(io, password)
+    elif encoding is Encoding.DER:
+        yield load_der(io.read() if not isinstance(io, bytes) else io, password)
     else:
         raise ValueError(f"Invalid encoding {encoding}")
 
